package happert;

import happert.UI.HapperMessages;
import happert.UI.IRender;
import happert.UI.KeyStateTracker;
import happert.gameobjects.DirectionEnum;

import happert.gameobjects.Player;
import happert.gameobjects.Happert;

import happert.playfield.Playfield;
import happert.playfield.PlayfieldBuilder;
import happert.playfield.RandomPlayfieldBuilder;


public class Game {

    static class TimeTracker {
        private long frequency;
        private long lastPulseTime;

        TimeTracker(long frequency) {
            assert frequency > 0 : "frequency shouldn't be zero.";
            this.frequency = frequency;
        }

        public void startTiming() {
            lastPulseTime = getCurrentTime();
        }

        public long timeTillNextPulse() {
            long playerWait = frequency - (getCurrentTime() - lastPulseTime);
            return playerWait;
        }

        public boolean isPulseTime() {
            return (getCurrentTime() - lastPulseTime) > frequency;
        }

        public void pulseCompleted() {
            lastPulseTime = getCurrentTime();
        }

        // testing seam
        protected long getCurrentTime() {
            return System.currentTimeMillis();
        }
    }

//    public enum PlayfieldBuilderEnum {
//        Random,
//        StringTemplate
//    }
//    private PlayfieldBuilderEnum builder = PlayfieldBuilderEnum.Random;

    private PlayfieldBuilder builder;

    private Playfield playfield;
    private Player player;
    private Happert happert;

    private KeyStateTracker keyStateTracker;

    private GameDifficultySettings gameDifficultySettings;

    private boolean stopped = true;
    private boolean loopActive = false;

    private IRender renderer;

    private static final long DELAY_PLAYER = 175;
    private TimeTracker playerTimer = new TimeTracker(DELAY_PLAYER);

    private static final long DELAY_HAPPERT = 675;
    private TimeTracker happertTimer = new TimeTracker(DELAY_HAPPERT);

    public Game() {
        gameDifficultySettings = new GameDifficultySettings();
        gameDifficultySettings.setDifficulty( GameDifficulty.Normal );
    }

    public void init() {

        RandomPlayfieldBuilder pfb = new RandomPlayfieldBuilder();
        // if the renderer is bound already let's check it's support width and height
        // (only supported by the random playfieldbuilder)
        if (renderer != null) {
            pfb.setCellDimensions(renderer.viewableCellWidth(), 
                                renderer.viewableCellHeight());
        }
        init(pfb);
    }

    public void init(PlayfieldBuilder pfb) {
        pfb.construct( this );
        playfield = pfb.getPlayfield();
        player = pfb.getPlayer();
        happert = pfb.getHappert();
        this.builder = pfb;
    }

    public Playfield getPlayfield() {
        return playfield;
    }

    /**
     * @return the player
     */
    public Player getPlayer() {
        return player;
    }

    /**
     * @return the happert
     */
    public Happert getHappert() {
        return happert;
    }

    public void attachKeyStateTracker(KeyStateTracker keyStateTracker) {
        this.keyStateTracker = keyStateTracker;
    }
    
    public void Start() {
        // set game settings
        setGameSettings();
        
        // put game loop on it's own thread
        new Thread(new Runnable() {
            public void run() {
                start();
            }
        }).start();
    }

    private void start() {
        assert keyStateTracker != null : "KeyStateTracker is not supposed to be null.";
        stopped = false;

        happert.huntFor(player);

        // reset timings
        playerTimer.startTiming();
        happertTimer.startTiming();

        System.out.println("Game started");
        
        // game loop
        boolean gameEnd = false;
        loopActive = true;
        try {
            do {
                // sleep while both objects are awaiting their pulse
                sleepTimeGap(playerTimer.timeTillNextPulse(),
                            happertTimer.timeTillNextPulse()
                        );

                // game tick

                // check player controls, see if the user wants to move
                if (playerTimer.isPulseTime()) {
                    DirectionEnum ks = keyStateTracker.getDirection();
                    // we won't waste our 'pulse' while the user is standing still;
                    // when he presses a key later, reponse will be instant
                    if (ks != DirectionEnum.None) {
                        player.move(ks);
                        playerTimer.pulseCompleted();
                    }
                }

                if (happertTimer.isPulseTime()) {
                    // move the happert
                    if ( happert.move() == false ) {
                        renderer.sendMessageToWindow(HapperMessages.WON);
                        Stop();
                    }
                    happertTimer.pulseCompleted();
                }

                // redraw screen
                renderer.Render();

                // check if we are in lose or play state
                if( ! gameIsValid() ) {
                    renderer.sendMessageToWindow(HapperMessages.LOOSE);
                    Stop();
                }
                // game end condition or player stopped game?
            } while(!gameEnd && !stopped);

        } finally {
            // make sure this flag gets cleared no matter what
            loopActive = false;
        }
        System.out.println("Game stopped");

    }

    public void Reset() {
        stopped = true;
        // wait untill game loop really quit running
        while (loopActive) {
            try {
                Thread.sleep(100);
            } catch (InterruptedException ex) {
                // ignore
            }
        }

        // TODO: maybe we should get a new instance of the builder?
        this.init(builder);
        renderer.Render();
    }

    public void RendererResize() {
        // while game isn't running and we are using randomplayfield
        if (stopped) {
            if (renderer != null && builder instanceof RandomPlayfieldBuilder) {
                // re-init
                init();
                renderer.Render();
            }
        }
    }

    protected void sleepTimeGap(long playerWait, long happertWait) {
        // if both are still waiting sleep a bit
        if (playerWait > 0 && happertWait > 0) {
            long wait = calculateTimeGap(playerWait, happertWait);
            // sleep a bit
            try {
                Thread.sleep(wait);
            } catch (InterruptedException ex) {
                // we don't care really: we have no other threads that could interrupt this one
            }
        }
    }

    public long calculateTimeGap( long playerWait, long happertWait ) {
        long wait = 0;
        if (playerWait > 0 && happertWait > 0) {
            if (playerWait < happertWait) {
                wait = playerWait;
            } else {
                wait = happertWait;
            }
        }

        return wait;
    }

    public void Stop() {
        stopped = true;
    }

    void setRenderer(IRender hp) {
        renderer = hp;
    }

    private boolean gameIsValid() {
        // Test if the happer is on the same location as the player
        if( happert.getLocation().equals( player.getLocation() ) ) {
            return false;
        }
        return true;
    }

    /**
     * Whether or not the game is running
     * @return
     */
    public boolean isRunning() {
        return loopActive;
    }
    /**
     * Get the current difficulty which is set
     * @return GameDifficultySettings The current game settings
     */
    public GameDifficultySettings getDifficultySettings() {
        return gameDifficultySettings;
    }

    /**
     * Set the game difficulty
     *
     * @param GameDifficultySettings gd
     */
    public void setDifficultySettings( GameDifficultySettings gd ) {
        gameDifficultySettings = gd;
    }

    /**
     * Set the Happer speed
     */
    private void setGameSettings() {
        // set Happer speed according settings
        happertTimer = new TimeTracker( gameDifficultySettings.getHapperSpeed() );
    }

}
