package happert.gameobjects;

import happert.playfield.Playfield;
import happert.playfield.Point;

public abstract class MovableGameObject extends GameObject {

    private Playfield field;

    public abstract boolean pushable();

    public MovableGameObject(Point startLocation, Playfield playfield) {
        super(startLocation);
        if (playfield == null) {
            throw new NullPointerException("Playfield can't be null");
        }
        this.field = playfield;
    }

    /**
     * Calculate the new location of the object
     * @param direction
     * @return
     */
    public Point calculateNewLocation(DirectionEnum direction) {
        assert direction != DirectionEnum.None : "Direction is set to none; doesn't lead to a new location.";
        Point oldLoc = this.getLocation();
        return oldLoc.moveDirection(direction, field.getWidth(), field.getHeight());
    }

    /**
     * Calculate a new location and check if it's a valid move. If so
     * perform the actual move
     *
     * @param direction
     * @return true when succesful, false when moving was not possible
     *
     */
    public boolean move(DirectionEnum direction) {
        if (direction == DirectionEnum.None) {
            return false;
        }

        Point newLoc = calculateNewLocation(direction);

        // direction doesn't yield a valid new location
        if (newLoc == null) {
            return false;
        }

        GameObject targetCellContent = field.getCellContents(newLoc);
        if (targetCellContent != null) {
            // target cell is not empty
            if (!resolveBlock(targetCellContent, direction)) {
                return false;
            }
        }

        moveInPlayfield(newLoc);
        setLocation(newLoc);
        return true;
    }

    /**
     * When running into some object when trying to move this method will be called
     * in an attempt to resolve the blockade.
     * The default implementation here tries to push objects forward if they are
     * pushable.
     *
     * @param targetCellContent the blocking GameObject
     * @param direction
     * @return true when succesful, false when moving was not possible
     *
     */
    protected boolean resolveBlock(GameObject targetCellContent, DirectionEnum direction) {
        // can the GameObject in target cell move ?
        if (targetCellContent instanceof MovableGameObject) {
            MovableGameObject moveTarget = (MovableGameObject) targetCellContent;
            // It does move but does it allow being pushed?
            if (!moveTarget.pushable()) {
                return false;
            }
            // Push it!
           return moveTarget.move(direction);
        }
        return false;
    }

    /**
     * Perform an actual move of the element
     * @param newLoc
     */
    private void moveInPlayfield(Point newLoc) {

        assert field.getCellContents(getLocation()).equals(this) :
                "Playfield doesn't hold 'this' object at the current location";

        field.setCellContents(getLocation(), null);
        field.setCellContents(newLoc, this);
    }

    /**
     * @return the field
     */
    public Playfield getPlayfield() {
        return field;
    }
}
