package happert.gameobjects.happertstrategies;

import happert.gameobjects.DirectionEnum;
import happert.playfield.Point;
import java.util.ArrayList;
import java.util.Collections;

import java.util.EmptyStackException;
import java.util.Stack;

class Dijkstra {
    
    private final int DEFAULT_WAIT = 1;
    
    private final ISearchField field;
    private final Point start;
    private final Point target;
    private int searchTime = 0;

    final ArrayList<Scout> waitingScouts = new ArrayList<Scout>();
    final DirectionEnum[][] scoutRegister;
    
    final Stack<DirectionEnum> routeStack = new Stack<DirectionEnum>();

    public Dijkstra(ISearchField field, Point start, Point target) {
        if (field == null) {
            throw new NullPointerException("field is null");
        }
        if (start == null) {
            throw new NullPointerException("start is null");
        }
        if (target == null) {
            throw new NullPointerException("target is null");
        }

        this.field = field;
        this.target = target;
        this.start = start;

        scoutRegister = new DirectionEnum[field.getWidth()][field.getHeight()];

    }

    boolean find() {
        
        // kickstart search by setting up mother scout
        Point loc = start;
        scoutRegister[loc.X()][loc.Y()] = DirectionEnum.None;
        Scout s = new Scout(0, loc, DirectionEnum.None);
        waitingScouts.add(s);

        boolean found = false;
        while (waitingScouts.size() > 0 && !found) {
            Scout scout = waitingScouts.get(0);
            // do an iteration
            // make searchTime the lowest time currently in the list
            searchTime = waitingScouts.get(0).getWait();
            while (scout.getWait() == searchTime) {
                // let it spawn children
                spawnScouts(scout.getPoint());
                // kill this scout
                waitingScouts.remove(scout);

                // grab next scout
                if (waitingScouts.size() == 0) break;
                
                scout = waitingScouts.get(0);
                // target hit with this scout?
                if (scout.getPoint().equals(target)) {
                    found = true;
                    break;
                }

            }
            // order list by waiting time
            Collections.sort(waitingScouts);
        }
        // reconstruct route from register pushing each step on the stack (FILO)
        // basically walking the route we found in reverse by inverting the stored vectors
        if (found) {
            Point routeLoc = target;

            do {
                DirectionEnum dir = scoutRegister[routeLoc.X()][routeLoc.Y()];
                routeStack.push(dir);
                routeLoc = routeLoc.moveDirection(dir.inverse(), field.getWidth(), field.getHeight());
            } while (!routeLoc.equals(start));
        }
        return found;
    }

    void spawnScouts(Point origin) {
        // loop over all directions and see if a move is possible in that direction
        // add a scout to the queue if so
        for (DirectionEnum dir : DirectionEnum.values()) {
            if (dir == DirectionEnum.None) {
                continue; // skip the None direction
            }
            Point scoutLoc = origin.moveDirection(dir, field.getWidth(), field.getHeight());
            if (scoutLoc != null) {
                // is it visited already?
                if (!visited(scoutLoc)) {
                    if (field.passable(scoutLoc) || scoutLoc.equals(target)) {
                        Scout s = new Scout(searchTime + DEFAULT_WAIT, scoutLoc, dir);
                        waitingScouts.add(s);
                        visit(scoutLoc, dir);
                    }
                }
            }
        }
    }

    DirectionEnum popStep() {
        try {
            DirectionEnum stackTop = routeStack.pop();
            return stackTop;
        } catch (EmptyStackException ex) {
            return null;
        }
    }

    private boolean visited(Point loc) {
        return scoutRegister[loc.X()][loc.Y()] != null;
    }

    private void visit(Point loc, DirectionEnum arrivedFrom) {
        assert arrivedFrom != DirectionEnum.None : "arrivedFrom can't be null";
        scoutRegister[loc.X()][loc.Y()] = arrivedFrom;
    }
}
