/*
 * This is a small playfieldbuilder that uses a string array to express a game
 * playfield. These string arrays can be used to express tests or to create
 * template based playfields.
 *
 * Example:
 * 
 * .....    Legend:
 * ..PO.    
 * .O#..    . = empty cell
 * ..H..    P = Player
 * .....    H = Happert
 *          O = Rock (movable)
 *          # = Granite (not movable)
 *
 * top left corner = (0, 0)
 * 
 */

package happert.playfield;

import happert.Game;
import happert.gameobjects.*;

public class StringTemplatePlayfieldBuilder extends PlayfieldBuilder {
    private String[] template = null;

    @Override
    public void construct(Game game) {
        if (template == null) {
            throw new NullPointerException("Please set a template first.");
        }
        construct(game, template);
    }

    private void construct(Game game, String[] template) {

        // all strings should have equal width
        int width = template[0].length();

        playfield = new Playfield(width, template.length);
        goBuilder = new GameObjectBuilder(playfield);

        for (int i = 0; i < template.length; i++) {
            if (template[i].length() != width) {
                throw new IllegalArgumentException("All string elements in template must be equal length");
            }

            // what sort of element is it?
            for (int tokenIndex = 0; tokenIndex < width; tokenIndex++) {
                Point loc = new Point(tokenIndex, i);
                GameObjectBuilder.GameObjectEnum objType;

                switch (template[i].charAt(tokenIndex)) {
                    case 'P' : // Player
                        if (player != null) {
                            throw new IllegalArgumentException("Player was already positioned, make sure the template only contains one.");
                        }
                        objType = GameObjectBuilder.GameObjectEnum.PLAYER;
                        break;
                    case 'H' :  // Happert
                        if (happert != null) {
                            throw new IllegalArgumentException("Happer was already positioned, make sure the template only contains one.");
                        }
                        objType = GameObjectBuilder.GameObjectEnum.HAPPERT;
                        break;
                    case '#' : // Granite
                        objType = GameObjectBuilder.GameObjectEnum.GRANITE;
                        break;
                    case 'O' : // Rock
                    case '0' : // Alternative char for Rock
                        objType = GameObjectBuilder.GameObjectEnum.ROCK;
                        break;
                    case '.' : // empty cell, do nothing
                        objType = null;
                        break;
                    default:
                        throw new IllegalArgumentException("Unkown template character " + template[i].charAt(tokenIndex) + " at " + loc);
                }

                if (objType != null) {
                    GameObject go = goBuilder.createGameObject(objType, loc);
                    playfield.setCellContents(loc, go);

                    // Need to track Player and Happert instances
                    if (objType == GameObjectBuilder.GameObjectEnum.PLAYER) {
                        player = (Player)go;
                    }
                    if (objType == GameObjectBuilder.GameObjectEnum.HAPPERT) {
                        happert = (Happert)go;
                    }
                }
            }
        }
        if (player == null || happert == null) {
            throw new IllegalArgumentException("Missing a required game element in playfield template, Happert of Player.");
        }
    }

    /**
     * @return the template
     */
    public String[] getTemplate() {
        return template;
    }

    /**
     * @param template the template to set
     */
    public void setTemplate(String[] template) {
        this.template = template;
    }

}
