package happert.gameobjects.happertstrategies;

import happert.gameobjects.DirectionEnum;
import happert.playfield.PlayfieldBuilder;
import happert.playfield.Point;
import happert.testutil.TestUtility;
import org.junit.Test;
import static org.junit.Assert.*;

public class DijkstraTest {

    private final Point aPoint = new Point(0,0);

    private class dummyField implements ISearchField {
        public int getWidth() {
            return 5;
        }
        public int getHeight() {
            return 5;
        }
        public boolean passable(Point loc) {
            throw new UnsupportedOperationException("Not supported yet.");
        }
    }

    @Test
    public void NullField_Dijkstra_ShouldThrowException() {
       
        boolean exc = false;
        try {
             Dijkstra d = new Dijkstra(null, aPoint, aPoint);
        } catch (Exception ex) {
            exc = true;
        }
        assertTrue(exc);
    }

    @Test
    public void NullStart_Dijkstra_ShouldThrowException() {
        ISearchField field = new dummyField();

        boolean exc = false;
        try {
            Dijkstra d = new Dijkstra(field, null, aPoint);
        } catch (Exception ex) {
            exc = true;
        }
        assertTrue(exc);
    }

    @Test
    public void NullTarget_Dijkstra_ShouldThrowException() {
        
        ISearchField field = new dummyField();

        boolean exc = false;
        try {
            Dijkstra d = new Dijkstra(field, aPoint, null);
        } catch (Exception ex) {
            exc = true;
        }
        assertTrue(exc);
    }

    @Test
    public void OpenField_onepass_spawnScouts() {
        // arrange
        String [] pft = {   ".....",
                            ".H..P",
                            "....."
                        };

        PlayfieldBuilder pfb = TestUtility.constructPlayfieldFromStringTemplate(pft);
        Point start = pfb.getHappert().getLocation();
        Point target = pfb.getPlayer().getLocation();
        
        Dijkstra d = new Dijkstra(new PlayfieldAdapter(pfb.getPlayfield()), start, target);
        
        // act
        d.spawnScouts(start);
        
        // assert
        assertEquals(4, d.waitingScouts.size());
    }

    @Test
    public void SemiOpenField_onepass_spawnScouts() {
        // arrange
        String [] pft = {   ".#...",
                            ".H#.P",
                            "....."
                        };

        PlayfieldBuilder pfb = TestUtility.constructPlayfieldFromStringTemplate(pft);
        Point start = pfb.getHappert().getLocation();
        Point target = pfb.getPlayer().getLocation();

        Dijkstra d = new Dijkstra(new PlayfieldAdapter(pfb.getPlayfield()), start, target);

        // act
        d.spawnScouts(start);

        // assert
        assertEquals(2, d.waitingScouts.size());
    }

    @Test
    public void BoxedInField_onepass_spawnScouts() {
        // arrange
        String [] pft = {   ".#...",
                            "#H#.P",
                            ".#..."
                        };

        PlayfieldBuilder pfb = TestUtility.constructPlayfieldFromStringTemplate(pft);
        Point start = pfb.getHappert().getLocation();
        Point target = pfb.getPlayer().getLocation();

        Dijkstra d = new Dijkstra(new PlayfieldAdapter(pfb.getPlayfield()), start, target);

        // act
        d.spawnScouts(start);

        // assert
        assertEquals(0, d.waitingScouts.size());
    }
    
    @Test
    public void BlockedField_onepass_spawnScouts_shouldBackOut() {
        // arrange
        String [] pft = {   ".##..",
                            ".H.#P",
                            ".##..",
                            "....."
                        };

        PlayfieldBuilder pfb = TestUtility.constructPlayfieldFromStringTemplate(pft);
        Point start = pfb.getHappert().getLocation();
        Point target = pfb.getPlayer().getLocation();

        Dijkstra d = new Dijkstra(new PlayfieldAdapter(pfb.getPlayfield()), start, target);

        // act
        d.spawnScouts(start);

        // assert
        assertEquals(2, d.waitingScouts.size());
        assertEquals(DirectionEnum.Left, d.waitingScouts.get(0).getDirection());
        assertEquals(DirectionEnum.Right, d.waitingScouts.get(1).getDirection());

        assertEquals(null, d.scoutRegister[1][1]);
        assertEquals(DirectionEnum.Left, d.scoutRegister[0][1]);
        assertEquals(DirectionEnum.Right, d.scoutRegister[2][1]);

    }
}
